import os, traceback
from PySide6 import QtCore, QtWidgets, QtGui
from PIL import Image, ImageQt, UnidentifiedImageError


class Interface_EnemyEdit(QtWidgets.QDialog):
    def __init__(self, enemy_screen, enemy_names):
        super().__init__()

        # variables
        self.enemy_screen = enemy_screen
        self.enemy_names = enemy_names
        
        # controls
        self.root = QtWidgets.QVBoxLayout()
        self.root_widget = QtWidgets.QWidget()
        self.setLayout(self.root)
        
        self.scroll = QtWidgets.QScrollArea()
        self.scroll.setWidgetResizable(True)
        self.scroll.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
        self.scroll.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.root.addWidget(self.scroll)
        
        self.scroll_vbox = QtWidgets.QVBoxLayout()
        self.scroll_vbox_widget = QtWidgets.QWidget()
        self.scroll_vbox_widget.setLayout(self.scroll_vbox)
        self.scroll.setWidget(self.scroll_vbox_widget)
        
        self.enemy_vbox = QtWidgets.QVBoxLayout()
        self.scroll_vbox.addLayout(self.enemy_vbox)
        self.scroll_vbox.addStretch() # vertical align to the top
        
        self.enemy_add_button = QtWidgets.QPushButton("Add enemy")
        self.root.addWidget(self.enemy_add_button)
        
        self.buttonbox = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel)
        self.buttonbox.accepted.connect(self.accept)
        self.buttonbox.rejected.connect(self.reject)
        self.root.addWidget(self.buttonbox)
        
        # init function calls
        for enemy in self.enemy_screen:
            enemy_widget = self.create_enemy_widget(enemy)
            self.enemy_vbox.addWidget(enemy_widget)
        
        # controller
        self.enemy_add_button.clicked.connect(self.add_enemy)


    def create_enemy_widget(self, enemy):
        # controls
        vbox = QtWidgets.QVBoxLayout()
        vbox_widget = QtWidgets.QWidget()
        vbox_widget.setLayout(vbox)
        vbox_widget.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        
        hbox_top = QtWidgets.QHBoxLayout()
        vbox.addLayout(hbox_top)
        
        respawn_label = QtWidgets.QLabel()
        respawn_label.setText("Respawn ID: ")
        hbox_top.addWidget(respawn_label)
        
        respawn_textbox = QtWidgets.QLineEdit()
        respawn_textbox.setText("{:02X}".format(enemy["respawn"]))
        hbox_top.addWidget(respawn_textbox)
        
        delete_button = QtWidgets.QPushButton("Delete")
        hbox_top.addWidget(delete_button)
        
        hbox_middle = QtWidgets.QHBoxLayout()
        vbox.addLayout(hbox_middle)
        
        name_label = QtWidgets.QLabel()
        name_label.setText("Enemy: ")
        hbox_middle.addWidget(name_label)
        
        name_combobox = QtWidgets.QComboBox()
        name_combobox.addItems(self.enemy_names)
        name_combobox_index = name_combobox.findText(enemy["name"], QtCore.Qt.MatchFixedString)
        if name_combobox_index < 0:
            raise Exception("enemy name '" + enemy["name"] + "' is not in name_combobox")
        name_combobox.setCurrentIndex(name_combobox_index)
        hbox_middle.addWidget(name_combobox)
        
        hbox_bottom = QtWidgets.QHBoxLayout()
        vbox.addLayout(hbox_bottom)
        
        x_label = QtWidgets.QLabel()
        x_label.setText("X: ")
        hbox_bottom.addWidget(x_label)
        
        x_textbox = QtWidgets.QLineEdit()
        x_textbox.setText("{:02X}".format(enemy["x"]))
        hbox_bottom.addWidget(x_textbox)
        
        y_label = QtWidgets.QLabel()
        y_label.setText("Y: ")
        hbox_bottom.addWidget(y_label)
        
        y_textbox = QtWidgets.QLineEdit()
        y_textbox.setText("{:02X}".format(enemy["y"]))
        hbox_bottom.addWidget(y_textbox)
        
        def save_widget_to_enemy_screen():
            widgets = [self.enemy_vbox.itemAt(i).widget() for i in range(self.enemy_vbox.count())]
            index = widgets.index(vbox_widget)
            enemy = self.enemy_screen[index]
            
            respawn = respawn_textbox.text()
            name = name_combobox.currentText()
            x = x_textbox.text()
            y = y_textbox.text()
            
            try:
                respawn = int(respawn, 16)
                x = int(x, 16)
                y = int(y, 16)
                
                assert respawn >= 0 and respawn < 0x80
                assert name in self.enemy_names
                assert x >= 0 and x < 256
                assert y >= 0 and y < 256
                
                enemy["respawn"] = respawn
                enemy["name"] = name
                enemy["x"] = x
                enemy["y"] = y
            except Exception as ex:
                messageBox = QtWidgets.QMessageBox(self)
                messageBox.setWindowTitle("Invalid property value")
                messageBox.setText(str(ex))
                messageBox.setDetailedText(str(traceback.format_exc()))
                messageBox.setIcon(QtWidgets.QMessageBox.Icon.Critical)
                messageBox.exec_()
            
        def delete_enemy_widget():
            widgets = [self.enemy_vbox.itemAt(i).widget() for i in range(self.enemy_vbox.count())]
            index = widgets.index(vbox_widget)
            vbox_widget.deleteLater()
            self.enemy_screen.pop(index)
            
        # controller
        respawn_textbox.textChanged.connect(save_widget_to_enemy_screen)
        name_combobox.currentIndexChanged.connect(save_widget_to_enemy_screen)
        x_textbox.textChanged.connect(save_widget_to_enemy_screen)
        y_textbox.textChanged.connect(save_widget_to_enemy_screen)
        delete_button.clicked.connect(delete_enemy_widget)
        
        return vbox_widget

    
    def add_enemy(self):
        enemy = {
            "respawn": 0,
            "name": self.enemy_names[0],
            "x": 0,
            "y": 0,
        }
        enemy_widget = self.create_enemy_widget(enemy)
        self.enemy_vbox.addWidget(enemy_widget)
        self.enemy_screen.append(enemy)
        
