from PIL import Image
import sys, os

# TODO: fix bug where a tileset that does not have any custom air tiles will have bad solidity

def main():
    if len(sys.argv) != 2:
        sys.exit("usage: " + sys.argv[0] + " <imagefile>")
    
    filename = os.path.splitext(sys.argv[1])[0]

    try:
        im = Image.open(sys.argv[1])
    except FileNotFoundError:
        sys.exit("file was not found")

    with im:
        im = im.convert('RGB')
        
        if im.size[0] % 16 != 0 or im.size[1] % 16 != 0:
            sys.exit("image dimensions should be multiples of 16, but are " + str(im.size))
        if im.size[0] / 16 * im.size[1] / 16 > 128:
            sys.exit("image should have at most 128 metatiles, but it has " + str(im.size[0] / 16 * im.size[1] / 16))

        im_metatiles, im_collision, im_airflag = im.split()

        if im_metatiles.getcolors() is None:
            sys.exit("image's metatiles should have 4 colors, but they have more than 256 (how did this happen?)")
        if len(im_metatiles.getcolors()) > 4:
            sys.exit("image's metatiles should have 4 colors, but they have " + str(len(im_metatiles.getcolors())))
        if len(im_metatiles.getcolors()) <= 1:
            sys.exit("image contains empty metatile data")

        if im_collision.getcolors() is None:
            sys.exit("image's collision should have 256 colors, but it has more than 256 (how did this happen?)")

        if im_airflag.getcolors() is None:
            sys.exit("image's airflag should have 4 colors, but it has more than 256 (how did this happen?)")
        if len(im_airflag.getcolors()) > 4:
            sys.exit("image's airflag should have 4 colors, but it has " + str(len(im_airflag.getcolors())))
        
        
        # get color indexes
        metatile_colors = list(map(lambda x: x[1], im_metatiles.getcolors()))
        metatile_colors.sort(reverse=True)
        metatile_colors.insert(0, metatile_colors.pop())
        

        # get tiles in unique array
        # get tileset map
        empty_tile = Image.new("L", (8, 8), color=metatile_colors[0])

        solidities = [0b000, -1, -1, 0b111]
        tiles_of_solidity = [[] for _ in solidities]
        tile_id_map_of_solidity = [[] for _ in solidities]
        collision_of_solidity = [[] for _ in solidities]

        for ym in range(0, im.size[1] // 8, 2):
            for tile_id_map in tile_id_map_of_solidity:
                tile_id_map.append([])
                tile_id_map.append([])
            for xm in range(0, im.size[0] // 8, 2):
                for y in range(ym, ym+2):
                    for x in range(xm, xm+2):
                        for tile_id_map in tile_id_map_of_solidity:
                            tile_id_map[y].append(None)

                        tile = im_metatiles.copy().crop((x*8, y*8, (x+1)*8, (y+1)*8))
                        solidity = im_airflag.getpixel((x*8, y*8)) >> 5
                        solidity_id = sum([int(c) for c in bin(solidity)[2:]]) # how many airflags are True
                        if solidities[solidity_id] == -1:
                            solidities[solidity_id] = solidity
                        elif solidities[solidity_id] != solidity:
                            sys.exit("image contains conflicting airflag solidities " + str(solidities[solidity_id]) + " and " + str(solidity))
                        tiles = tiles_of_solidity[solidity_id]

                        tile_id_map = tile_id_map_of_solidity[solidity_id]
                        collision = collision_of_solidity[solidity_id]

                        if list(empty_tile.getdata()) == list(tile.getdata()) and solidity_id == len(solidities)-1 and im_collision.getpixel((x*8, y*8)) == 0x00:
                            continue

                        tile_should_be_added = True
                        for t_id, t in enumerate(tiles):
                            if list(t.getdata()) == list(tile.getdata()):
                                tile_should_be_added = False
                                tile_id_map[y][x] = t_id
                                break
                        if tile_should_be_added:
                            tiles.append(tile)
                            tile_id_map[y][x] = len(tiles)-1
                            collision.append(bytes([im_collision.getpixel((x*8, y*8))]))

        combined_tiles = sum(tiles_of_solidity, [])
        #combined_tiles += [empty_tile] * (128 - len(combined_tiles))

        if len(combined_tiles) > 128:
            sys.exit("image's metatiles should be composed from at most 128 tiles, but they are composed from " + len(combined_tiles))

        combined_collision = sum(collision_of_solidity, [])
        combined_collision += [bytes([0x00])] * (256 - len(combined_collision))

        combined_tile_id_map = []
        for y in range(im.size[1] // 8):
            combined_tile_id_map.append([])
            for x in range(im.size[0] // 8):
                tile_id = 255
                tiles_len = 0
                for solidity_id, tile_id_map in enumerate(tile_id_map_of_solidity):
                    if tile_id_map[y][x] != None:
                        tile_id = tile_id_map[y][x] + tiles_len
                        break
                    else:
                        tiles_len += len(tiles_of_solidity[solidity_id])
                combined_tile_id_map[y].append(tile_id)

        combined_metatiles = []
        for ym in range(0, im.size[1] // 8, 2):
            for xm in range(0, im.size[0] // 8, 2):
                for y in range(ym, ym+2):
                    for x in range(xm, xm+2):
                        combined_metatiles.append(combined_tile_id_map[y][x].to_bytes(1, byteorder='big'))

        combined_metatiles += [bytes([0xff])] * (512 - len(combined_metatiles))


        # write tileset map to asm
        with open(filename + "_metatiles.asm", "w") as asmfile:
            asmfile.write("; " + os.path.basename(filename) + "_metatiles.asm generated by " + os.path.basename(sys.argv[0]) + "\n")
            asmfile.write("    db ")
            for k, v in enumerate(combined_metatiles):
                asmfile.write("$" + v.hex())
                if (k+1) % 16 == 0:
                    asmfile.write("\n")
                    if k+1 < len(combined_metatiles):
                        asmfile.write("    db ")
                else:
                    asmfile.write(", ")


        # write tile collision to asm
        with open(filename + "_collision.asm", "w") as asmfile:
            asmfile.write("; " + os.path.basename(filename) + "_collision.asm generated by " + os.path.basename(sys.argv[0]) + "\n")
            asmfile.write("    db ")
            for k, v in enumerate(combined_collision):
                asmfile.write("$" + v.hex())
                if (k+1) % 16 == 0:
                    asmfile.write("\n")
                    if k+1 < len(combined_collision):
                        asmfile.write("    db ")
                else:
                    asmfile.write(", ")


        # write tiles to chr
        with open(filename + ".chr", "wb") as chrfile:
            for tile in combined_tiles:
                for y in range(8):
                    byte1 = bytes([0x00])
                    byte2 = bytes([0x00])
                    for x in range(8):
                        pixel = tile.getpixel((x, y))
                        pixel_id = metatile_colors.index(pixel)
                        byte1 = bytes([int.from_bytes(byte1, byteorder='big') + (0 if (pixel_id & 0x1) == 0 else 0x80 >> x)])
                        byte2 = bytes([int.from_bytes(byte2, byteorder='big') + (0 if (pixel_id & 0x2) == 0 else 0x80 >> x)])
                    chrfile.write(byte1)
                    chrfile.write(byte2)
        
        # write tile solidity to asm
        solidity_threshold = 0
        solidity_old = 0b000
        solidity_flags = [0b100, 0b010, 0b001]
        solidity_names = ["samus", "enemy", "projectile"]
        solidity_thresholds = [0 for _ in solidity_names]
        solidities_to_set = []
        
        for solidity_id, solidity in enumerate(solidities):
            solidity_len = len(tiles_of_solidity[solidity_id])
            if solidity_len > 0:
                for id_to_set in solidities_to_set:
                    solidity_difference = solidity - solidity_old
                    if solidity_difference <= 0:
                        sys.exit("next solidity (" + "${0:03b}".format(solidity_old) + ") should be greater than old solidity (" + "${0:03b}".format(solidity) + ")")
                    for flag_id, flag in enumerate(solidity_flags):
                        if flag <= solidity_difference:
                            break
                    solidity_thresholds[flag_id] = solidity_threshold
                    solidity_old += flag
                solidities_to_set = []
            solidity_threshold += solidity_len
            solidities_to_set.append(solidity_id)
        
        
        with open(filename + "_solidity.asm", "w") as asmfile:
            asmfile.write("; " + os.path.basename(filename) + "_solidity.asm generated by " + os.path.basename(sys.argv[0]) + "\n")
            asmfile.write("; " + ", ".join(solidity_names) + "\n")
            asmfile.write("    db " + ", ".join(["${0:02X}".format(threshold) for threshold in solidity_thresholds]))
        


if __name__ == "__main__":
    main()


